"""Artistools - spectra related functions."""

from artistools.spectra import plotspectra
from artistools.spectra import writespectra
from artistools.spectra.plotspectra import main as plot
from artistools.spectra.spectra import get_exspec_bins
from artistools.spectra.spectra import get_flux_contributions
from artistools.spectra.spectra import get_flux_contributions_from_packets
from artistools.spectra.spectra import get_from_packets
from artistools.spectra.spectra import get_reference_spectrum
from artistools.spectra.spectra import get_specpol_data
from artistools.spectra.spectra import get_spectrum
from artistools.spectra.spectra import get_spectrum_at_time
from artistools.spectra.spectra import get_vspecpol_data
from artistools.spectra.spectra import get_vspecpol_spectrum
from artistools.spectra.spectra import make_averaged_vspecfiles
from artistools.spectra.spectra import make_virtual_spectra_summed_file
from artistools.spectra.spectra import print_integrated_flux
from artistools.spectra.spectra import read_spec_res
from artistools.spectra.spectra import sort_and_reduce_flux_contribution_list
from artistools.spectra.spectra import stackspectra
from artistools.spectra.spectra import timeshift_fluxscale_co56law
from artistools.spectra.writespectra import write_flambda_spectra
